#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This collection of function includes all necessary processing to georeference images of the IR camera, the ice drift correction of the coordinates, and the gridding of all pixels to an equidistant grid of a given grid size (we used 1 m and 5 m). To get georeference, ice dirft corrected and girdded data the main function 'gridding_images()' has to be called.

@author: Marcus Huntemann, Linda Thielke
"""

# Required modules
import pyproj
import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
from netCDF4 import Dataset

import image_data_corrections as idc
from metadata_heli_flights import get_image_cutoff_flight

# Global parameters
WIDTH_ANGLE_CAM = 56.1 
HEIGHT_ANGLE_CAM = 43.6

PIXEL_WIDTH = 640
PIXEL_HEIGHT = 480


def get_mss(lat, lon, mss_nc_file):
    """
    Loading Mean sea surface height data. Calculating values along the flight track.
    Input
        - lon, lat: numpy.ndarray
        - mss_nc_file: str
    Output
        - mss_ts: list
    """
    data_mss = Dataset(mss_nc_file)
    mss = data_mss.variables['mss'][:]
    lat_mss=data_mss.variables['lat'][:]
    lon_mss=data_mss.variables['lon'][:]
    
    # getting index of closest pixel of mss data to flight track
    mss_ts = []
    for i in range(len(lon)):
        ix_lat = np.abs(lat_mss-lat[i]).argmin()
        ix_lon = np.abs(lon_mss-lon[i]).argmin()
        mss_ts.append(mss[ix_lat,ix_lon])
    
    return mss_ts


def get_pixel_distances(angle_width, angle_height, altitude, roll, pitch):
    """
    Calculates the distances for each pixel of the taken pictures.
    Based on flight altitude and angle of the camera.
    Includes pitch and roll angle of the Applanix as well as a scale factor
    fot the data in between center and edges.
    Input:
        -angle_width: Float
        -angle_height: Float
        -altitude: Float
        -roll: list
        -pitch: list
    Output:
        -points_width: np.ndarray
        -points_height: np.ndarray
    """
    
    # height angle with pitch correction
    angle_height_array = np.arange(-angle_height/2, angle_height/2, 
                                   angle_height/PIXEL_HEIGHT)
    angle_height_array = angle_height_array - pitch
    # width angle with roll correction
    angle_width_array = np.arange(-angle_width/2, angle_width/2, 
                                  angle_width/PIXEL_WIDTH)
    angle_width_array = angle_width_array - roll
    
    temp_arr= np.asarray([1]*PIXEL_HEIGHT) # dummy array to get 2D-array
    
    # distance to zenit in meter
    points_width = altitude * (np.tan(np.deg2rad(angle_width_array[np.newaxis,:]))
            / np.cos(np.deg2rad(angle_height_array[:,np.newaxis])))
    points_height = altitude * np.tan(np.deg2rad(angle_height_array[:, np.newaxis])) / np.cos(np.deg2rad(angle_width_array[np.newaxis,:]))
    
    return points_width, points_height


def get_rotated_coords(lon, lat, width_distance, height_distance,heading):
    """
    Calculates the corrdinates for the rotated positions of the data.
    An offset of the x and y component is computed in the rotated projection.
    -input:
        lon: float
        lat: float
        widths_distance: list
        height_distance: list
        heading: float
    -output:
        lon_proj: numpy.ndarray
        lat_proj: numpy.ndarray
    """
    
    head=heading-lon #adjust heading by longitude for orientation
    # values to range [-180,180]
    if head<-180: 
        head+=360
    elif head>180: 
        head-=360
    else:
        pass
    
    # Projection, rotation (lon) corrected by heading
    proj_dict = dict(proj='stere', 
                     lat_0=lat,
                     lat_ts=lat,
                     lon_0=-head,
                     k=1,
                     x_0=0,
                     y_0=0,
                     a=6378137, 
                     b=6378137,
                     units='m')
    
    # the x,y position for the turn by heading 
    p_base = pyproj.Proj(proj_dict)
    x_new, y_new = p_base(lon, lat) # for offset of projection
    
    # sustract the new position as offset to the original position
    # attention: negative heading for right orientation
    proj_dict_new = dict(proj='stere', 
                         lat_0=lat,
                         lat_ts=lat,
                         lon_0=-head,
                         k=1,
                         x_0=-x_new,
                         y_0=-y_new,
                         a=6378137, 
                         b=6378137,
                         units='m')
    
    # new projection with right relation to the north
    p = pyproj.Proj(proj_dict_new)

    # positions as lon and lat 
    lon_proj, lat_proj = p(width_distance,height_distance,inverse=True)

    return lon_proj, lat_proj, p 


def make_driftcor_fn(targettime, day=None):
    """
    Generates a drift correction including rotation using a projection 
    with coordinates in the center of the map.
    Input
        - targettime: float (target time of the grid (in seconds since daystart)
        - day: str
    Output
        - di1x,di1y: nterpolators for x and y
        - p2: projection instance converting lat lon to local coordinates (in meters) for the target time
        - FN_rot: bool
        - lon_rot: float
        - lon_interp: float
        - N_off: float
    """

    # RV Polarstern's coordinates as reference
    file='additional_data/mosaic_PS_gps-weather_dashboard_10min_processed.csv'
    d1 = pd.read_csv(file, index_col='datetime')
    d1 = d1[['longitude', 'latitude', 'heading']] 
    d1 = d1.rename(columns={'latitude': 'lat', 'longitude': 'lon'})
    d1.index = pd.to_datetime(d1.index)
    d1=d1[day]
    ixd = d1.index
    d1s=ixd.hour*60*60+ixd.minute*60+ixd.second 
    di1h=interp1d(d1s,d1.heading)

    # Reference time in another format
    targettime = (pd.to_datetime(targettime)-pd.to_datetime(targettime[:10])).seconds
    print('targettime: ', targettime)
    
    # Create initial coordinate system based on start coordinates of that day
    proj_dict = dict(proj='stere', 
                     lat_0=d1.lat[0],
                     lat_ts=d1.lat[0],
                     lon_0=d1.lon[0],
                     k=1,
                     x_0=0,
                     y_0=0,
                     a=6378137, 
                     b=6378137,
                     units='m')
    p=pyproj.Proj(proj_dict)
    
    # Convert to local grid coordinates
    d1x,d1y=p(d1.lon.values,d1.lat.values)

    # Preliminary interpolators for x and y of AIS Base stations over time
    di1x=interp1d(d1s,d1x)
    di1y=interp1d(d1s,d1y)

    #interpolated lat lon on x,y grid for better accuracy and working at the pole
    lon_interp,lat_interp=p(di1x(targettime),di1y(targettime),inverse=True)
    print('interpolated center coordinates: ', lon_interp,lat_interp)
    
    # Rotating of orientation to Reference
    lon_rot=lon_interp-di1h(targettime)+90
    N_off=di1h(targettime)+90
    if lon_rot<-180: 
        lon_rot+=360
    elif lon_rot>180: 
        lon_rot-=360
    else:
        pass
    FN_rot=True
    if np.isnan(lon_rot):
        print('No heading correction.')
        lon_rot=lon_interp
        FN_rot=False
        
    print('true heading: ', di1h(targettime))
    print('lon_rotated: ', lon_rot)
    
    # Rotated projections
    proj_dict = dict(proj='stere', 
                     lat_0=lat_interp,
                     lat_ts=lat_interp,
                     lon_0=lon_rot,
                     k=1,
                     x_0=0,
                     y_0=0,
                     a=6378137, 
                     b=6378137,
                     units='m')
    
    p_tmp=pyproj.Proj(proj_dict)
    x_tmp, y_tmp = p_tmp(lon_interp, lat_interp)
    print('x/y offset: ', x_tmp, y_tmp)    
    
    # Final projection including offset
    proj_dict = dict(proj='stere', 
                     lat_0=lat_interp,
                     lat_ts=lat_interp,
                     lon_0=lon_rot,
                     k=1,
                     x_0=-x_tmp,
                     y_0=-y_tmp,
                     a=6378137, 
                     b=6378137,
                     units='m')
    p2=pyproj.Proj(proj_dict)

    print(p2(lon_interp, lat_interp)) # for checking origin (should be 0,0)

    # Convert to local grid coordinaes
    d1x,d1y=p2(d1.lon.values,d1.lat.values)

    # Interpolators for x and y over time
    di1x=interp1d(d1s,d1x)
    di1y=interp1d(d1s,d1y)

    return (di1x,di1y),p2, FN_rot, lon_rot, lon_interp, N_off


def get_driftcor_ll(base_interpolators,lat,lon,targettime,intertime,p2):
    """
    Returns the lat lon values for an observation done at some point in time
    gets the drift offset which is lat, lon time dependent based on the 
    reference coordinates of RV Polarstern and targettime.
    Input
        - base_interpolators: list or tuple of the interpolators for the reference position
        - lat: (latitude of location to warp)
        - lon: (longitude of location to warp)
        - targettime: int (the time where the coordinate system is being warped to)
        - intertime: int (the actual time of the lat, lon values)
        - p2: pyproj instance of target grid (at intertime)
    Output 
        - p2(xn,yn,inverse=True): projected coordinates
    """
    
    di1x,di1y=base_interpolators 
    
    # movement, since position at targettime is (0,0) the position at intertime is the displacement
    dx1=di1x(intertime) 
    dy1=di1y(intertime)
    
    x,y=p2(lon,lat) #x,y in projection coordinates
    
    #apply drift in x and y direction
    xn=x-dx1 
    yn=y-dy1
    
    # rs=(xn**2+yn**2)**0.5 # distance from origin
    # print("distance from 0,0 point:", rs)
    # phis=np.arctan2(yn,xn) # angle from origin
    # print("rotation angle: ",phis)
    
    return p2(xn,yn,inverse=True)


def get_drift_coords(lon,lat,time,proclist,date,targettime):
    """
    Calculate drift corrected coordinates from image coordinates.
    Input:
        - lon,lat,time: pandas.DataFrame
        - proclist: list
        - date: str
        - targettime: 
    Output:
        - lons: array
        - lats: array
        - p: Projection
        - FN_rot: float
        - lon_rot: float
        - N_off: bool
    """
    
    base_interpolators, p, FN_rot, lon_rot, lon_interp, N_off = make_driftcor_fn(targettime, day=date)

    rlon,rlat=get_driftcor_ll(base_interpolators, lat[proclist], lon[proclist],
                                     targettime, time[proclist], p)

    lons = rlon
    lats = rlat
        
    return lons, lats, p, FN_rot, lon_rot, lon_interp, N_off


def nuc_correction(var, i):
    """   
    detecting NUC and remove frame before and NUC frames for MOSAiC default of 4fps
    the idea is the following when nuc apears this freezen the screen
    image for some time. That does not mean that with 4fps one complete
    frame is frozen since the length of the nuc is not always the same.
    This decets if within the last and next image a couple for lines close
    to the previous one are the same or very similar. If this is the case,
    either the helicopter was very slow (unlikely since the groundspeed for
    laser scanner flights is 90knots) or there was a NUC if there
    was a NUC in any of the frames, these should be skipped because the
    location from the INS data will be off by some pixels which will be
    visible in the mosaic.
    Input
        - var: numpy.MaskedArray
        - i: int
    Output
        - i: int
    """
    
    vm=var[i-1][225:255,:]
    v1=var[i][225:255,:]
    v2=var[i+1][225:255,:]
    nuctest_forward=np.abs(v1-v2).mean(axis=1).min()
    nuctest_backward=np.abs(v1-vm).mean(axis=1).min()
    # print(len(proclist),nuctest_forward,nuctest_backward)
    # print (i,nuctest)
    if nuctest_forward==0.0: 
        #a difference of 0 means that the frames must
        #be identical, this means thta the second frame i.e., i+1 is still
        #during the NUC, thus is should be skipped as well
        i+=2
    elif nuctest_forward<v1.std(): 
        #if the central line is within 
        #+/- 15 lines of the image (+noise) then NUC 
        #ended between i and i+1 and thus i+1 is not effected by the NUC
        i+=1
    elif nuctest_backward<v1.std():
        #I forgot why to do this probably there was a reason
        i+=1
        
    return i


def gridding_images(loadfile,grid_size, file_mss):
    """
    Main function: georeferencing and gridding + save to NetCDF file:
    Input
        - grid_size: int
        - loadfile: str
        - file_mss: str (file name of NetCDF of MSS from DTU, online available; https://ftp.space.dtu.dk/pub/DTU21/1_MIN/)
    Output
        - target, time_arr, roll_arr, pitch_arr: numpy.ndarray
    """
    
    # Data loading
    print('Loading image data:')
    data=Dataset(loadfile)
    devops=data.Device_Operation
    fid=data.Flight_ID
    
    lon_img=data.variables['lon_georef'][:]
    lat_img=data.variables['lat_georef'][:]
    lon_nadir=data.variables['lon_nadir'][:]
    lat_nadir=data.variables['lat_nadir'][:]
    temps=data.variables['Ts'][:]
    time=data.variables['time'][:]
    heading=data.variables['head'][:]
    altitude=data.variables['alti'][:]
    roll=data.variables['roll'][:]
    pitch=data.variables['pitch'][:]    
    
    # Image cut
    edgecut=100
    assert edgecut<480
    bi=np.s_[edgecut:-1-edgecut,:] #index for var respecting edgecut
    
    # Index list 
    tstart=0
    tstop=len(temps)
    
    cutoff = get_image_cutoff_flight()
    proclist=list(range(cutoff[devops][0]+tstart,
                        tstop-cutoff[devops][1],1))     
    
    # Targettime
    flight_times = pd.read_csv('additional_data/mosaic-heli-reftimes.csv',index_col='devops')
    ft = flight_times.loc[devops]
    targettime=ft.reftime
    print('Targettime: ', targettime)
    date= fid[:4]+'-'+fid[4:6]+'-'+fid[6:8] # Target day
    
    print('Calculating drift correction:') # output drift correction
    lons_dc, lats_dc, p, FN_rot, lon_rot, lon_interp, N_off = get_drift_coords(lon_nadir.data,lat_nadir.data,time.data,proclist,date,targettime)

    # Area settings
    xt,yt=p(lons_dc,lats_dc)
    xmin=np.min(xt)
    xmax=np.max(xt)
    ymin=np.min(yt)
    ymax=np.max(yt)

    x_delta=2000
    y_delta=2000
    xy_edges = np.round([xmin-x_delta,ymin-y_delta,xmax+x_delta,ymax+y_delta],-3)
    x_range=np.arange(xy_edges[0],xy_edges[2])
    y_range=np.arange(xy_edges[1],xy_edges[3])
    dx = np.abs(xy_edges[2] - xy_edges[0]) # grid size in x-direction
    dy = np.abs(xy_edges[3] - xy_edges[1]) # grid size in y-direction
    print('Grid size (x,y): ',dx,dy)

    # array allocation
    nxgrids=int(dx/grid_size)
    nygrids=int(dy/grid_size)
    
    Ts_arr=np.ones([nxgrids,nygrids],dtype=np.float32)*np.nan
    time_arr=np.ones([nxgrids,nygrids],dtype=np.float32)*np.nan
    roll_arr=np.ones([nxgrids,nygrids],dtype=np.float32)*np.nan
    pitch_arr=np.ones([nxgrids,nygrids],dtype=np.float32)*np.nan
    
    # creating coordinate grid
    x_grid = np.round(np.linspace(xy_edges[0],xy_edges[2],nxgrids))
    y_grid = np.round(np.linspace(xy_edges[1],xy_edges[3],nygrids))
    xv,yv = np.meshgrid(x_grid, y_grid)  
    lon_grid, lat_grid = p(xv,yv, inverse=True)
        
    gradient_array = idc.get_image_gradient_correction(temps) # correction array for images
    
    # mean sea surface hight along flight track for height correction
    mss = get_mss(lon_nadir.data,lat_nadir.data, file_mss) 
    
    print('Number of images for gridding: ', len(proclist))
    
    print('Gridding of images (note: this takes a few minutes):')
    while len(proclist)>2:
        ix=proclist.pop(0)
        i = nuc_correction(temps, ix) # adjust index for NUC occurences
        if np.abs(roll[i])<45: # limited incidence angle to avoid large distortions
            temp = temps[i]-gradient_array # correction of image
            T = np.asarray(temp)
            
            # distance of pixel from center point
            width_distance, height_distance = get_pixel_distances(WIDTH_ANGLE_CAM, 
                                                                  HEIGHT_ANGLE_CAM,
                                                                  altitude.data[i]-mss[i], 
                                                                  roll.data[i], 
                                                                  pitch.data[i])
            # ice drift, georeferenced coordinates of image pixel
            new_lon, new_lat, p_img = get_rotated_coords(lons_dc[i], 
                                                         lats_dc[i],
                                                         width_distance, 
                                                         height_distance, 
                                                         heading.data[i])

            try:
                # pixel to appropriate grid cell
                y,x=p(new_lon, new_lat)
                xind=x_range.searchsorted(y)[bi]
                yind=y_range.searchsorted(x)[bi]
                indind=xind<dx
                indind&=yind<dy 
                indind&=xind>0
                indind&=yind>0

                xind=(xind/grid_size).astype(int) # due to res > 1
                yind=(yind/grid_size).astype(int) 
                
                # Gridding of values
                Ts_arr[xind[indind],yind[indind]]=T[bi][indind].astype(np.float32)
                time_arr[xind[indind],yind[indind]] = time[i] 
                roll_arr[xind[indind],yind[indind]] = roll[i] 
                pitch_arr[xind[indind],yind[indind]] = pitch[i] 
            except:
                print('Image not gridded.')
                
    return Ts_arr, time_arr, roll_arr, pitch_arr, lon_grid, lat_grid


if __name__ == "__main__":
    """
    """
#    # --- call of main function:
#    Ts_map, time_map, roll_map, pitch_map, lon_map, lat_map = gridding_images(file,resolution) 