#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
In this script, we present function to read and display the gridded data. Also standard statsistical values can be caluclated.

@author: Linda Thielke
"""

# Required modules
from netCDF4 import Dataset
import matplotlib
import matplotlib.colors as clr
import matplotlib.pyplot as plt
import numpy as np

from pyproj import Proj

import cartopy.crs as ccrs
import cartopy.feature as cfeature

# global setting for figure font
font = {'size': 16}
matplotlib.rc('font', **font)

def read_map_nc(loadfile):
    """
    Data reading of gridded data from NetCDF file with netCDF4.Dataset.
    The current output is limited to the relative coordinates and the 
    surface temperature and can be extended indvidually.
    
    Input
        - loadfile: str
    Output
        - xc,yc,Ts: numpy.ma.core.MaskedArrays
        - FID: str
    """

    data=Dataset(loadfile)
    
    # Flight identification
    devops=data.Device_operation
    FID=data.Flight_ID
    
    print('Targettime: ', data.Targettime)
   
    grid_size=int(data.Grid_size[:-2]) # grid resolution in meter
    
    lon=data.variables['lon'][:]
    lat=data.variables['lat'][:]
    
    xc=data.variables['x'][:]
    yc=data.variables['y'][:]
    
    Ts=data.variables['Ts'][:]
    Ts_fixed=data.variables['Ts_fixed'][:]
    
    time=data.variables['time'][:]
    
    roll=data.variables['roll'][:]
    pitch=data.variables['pitch'][:]

    print('Projection information:')
    print(data.variables['stereographic'])
    
    return xc, yc, lon, lat, time, Ts, FID


def get_Ts_colormap():
    """
    Individually created colormap for surface temperature. Note that there are 
    high contrast in the middle of colormap. This maps os suited best to illustrate 
    the temperature sturcture of the whole temperature range of one flight. 
    For narrower temperature ranges we recomment 'plasma' from matplotlib 
    (matplotlib.pyplot.get_cmap('plasma')).
    Input
        - None
    Output
        - cmap_TIR: matplotlib.colors.LinearSegmentedColormap
    """
    
    cmap_TIR = clr.LinearSegmentedColormap.from_list('custom', 
    ['#004d99','#0080ff','#80bfff','#cce6ff','#ffe6b3','#e69900','#ff531a','#992600'],N=256)
    
    return cmap_TIR


def plotting_map_rc(x,y,data,cmap,FID,savefile=None):
    """
    This is a simple plotting routine to diplay a map, with the adjusted figure size,
    in relative coordinates. Labels might need to be adjusted.
    Input
        - x: numpy.ma.core.MaskedArrays
        - y: numpy.ma.core.MaskedArrays
        - data: numpy.ma.core.MaskedArrays
        - cmap: matplotlib.colors.LinearSegmentedColormap
        - savepath: str (optional)
        - savefile: str (optional)
    Output
        - Figure will be created
    """
    
    # Figure size
    xsize=15
    ysize=(xsize-1)/(int(len(data[0])/len(data)))

    plt.figure(figsize=(xsize,ysize))
    plt.pcolormesh(x,y,data, cmap=cmap,
                   vmin=np.round(np.nanpercentile(data,1)),
                   vmax=np.round(np.nanpercentile(data,99)))
    plt.colorbar(label='var / dim ')
    plt.xlabel('distance in x-direction / m')
    plt.ylabel('distance in y-direction / m')
    
    plt.grid()
    plt.tight_layout()
    plt.title(FID)
    
    if savefile==None:
        print('No path and/or filename. Figure will be not save.')
        plt.show()
    else:
        plt.savefig(savepath+savefile)
        plt.close()
        
        
def mapping_cartopy(lon,lat,data,lon_center,lat_center,cmap,FID,savefile=None):
    """
    This is a plotting routine using cartopy to diplay a map, with the adjusted figure size. 
    Labels might need to be adjusted.
    Input
        - lon: numpy.ma.core.MaskedArrays
        - lat: numpy.ma.core.MaskedArrays
        - data: numpy.ma.core.MaskedArrays
        - lon_center: float
        - lat_center: float
        - cmap: matplotlib.colors.LinearSegmentedColormap
        - savepath: str (optional)
        - savefile: str (optional)
    Output
        - Figure will be created
    """

    # Figrue size
    xsize=15
    ysize=(xsize-1)/(int(len(data[0])/len(data)))
    
    fig=plt.figure(figsize=(xsize,ysize))
    # Projections
    stere=ccrs.Stereographic(central_latitude=lat_center, 
                             central_longitude=lon_center)
    ax1=plt.subplot(111, projection=stere)
    
    # Set background color
    ax1.set_facecolor('lightgray')
    cm = cmap
    cm.set_bad('lightgray')
    
    c = ax1.pcolormesh(lon, lat, data, cmap=cmap,
                       vmin=np.round(np.nanpercentile(data,1)),
                       vmax=np.round(np.nanpercentile(data,99)),
                       transform=ccrs.PlateCarree())
    plt.plot(lon_center,lat_center,'kd')
    
    fig.colorbar(c, ax=ax1, fraction=0.039, label='var / dim', extend='both')
    
    # Scale
    proj_dict = dict(proj='stere', lat_0=lat_center, 
                     lat_ts=lat_center, lon_0=lon_center,
                     k=1, x_0=0, y_0=0,
                     a=6378137, b=6378137, units='m')
    m=Proj(proj_dict)
    x,y = m(lon,lat)
    
    xmin,xmax,ymin,ymax=(np.min(x),np.max(x),np.min(y),np.max(y))
    ax1.plot([xmin+1000,xmin+3000], [ymin+100, ymin+100], transform=stere, color='k',
            linewidth=1, marker='|')
    ax1.text(xmin+1700,ymin+200, '2 km', fontsize=12)
    ax1.text(xmin+500, ymin+100, u'\u25B2\nN', transform=stere,
            horizontalalignment='center', verticalalignment='bottom')
    
    plt.grid()
    plt.tight_layout()
    plt.title(FID)
    
    if savefile==None:
        print('No path and/or filename. Figure will be not save.')
        plt.show()
    else:
        plt.savefig(savefile)
        plt.close()

        
# #######################################
# Basemap not included in the environment
# #######################################
# def mapping_basemap(lon,lat,data,lon_center,lat_center,cmap,FID,savefile=None):
#     """
#     This is a plotting routine using basemap to diplay a map, with the adjusted figure size. 
#     Labels might need to be adjusted.
#     Input
#         - lon: numpy.ma.core.MaskedArrays
#         - lat: numpy.ma.core.MaskedArrays
#         - data: numpy.ma.core.MaskedArrays
#         - lon_center: float
#         - lat_center: float
#         - cmap: matplotlib.colors.LinearSegmentedColormap
#         - savefile: str (optional)
#     Output
#         - Figure will be created
#     """
#     from mpl_toolkits.basemap import Basemap 
#     proj_dict = dict(proj='stere', lat_0=lat_center, 
#                      lat_ts=lat_center, lon_0=lon_center,
#                      k=1, x_0=0, y_0=0,
#                      a=6378137, b=6378137, units='m')
#     m=Proj(proj_dict)
#     x,y = m(lon,lat)
#     xmin,xmax,ymin,ymax=(np.min(xc),np.max(xc),np.min(yc),np.max(yc))
    
#     m = Basemap(width=xmax-xmin+5000,height=ymax-ymin+5000,
#                 resolution='l',projection='stere',
#                 lat_ts=lat_center,lat_0=lat_center,lon_0=lon_center)
    
#     m.drawcoastlines()
#     m.drawparallels(np.arange(60.,90.,2.))
#     m.drawmeridians(np.arange(-180.,181.,20.))
    
#     xx,yy = m(lon,lat)
#     m.pcolormesh(xx,yy,data,cmap=cmap,
#                  vmin=np.round(np.nanpercentile(data,1)),
#                  vmax=np.round(np.nanpercentile(data,99)))
    
#     plt.grid()
#     plt.tight_layout()
#     plt.title(FID)
    
#     if savefile==None:
#         print('No path and/or filename. Figure will be not save.')
#         plt.show()
#     else:
#         plt.savefig(savefile)
#         plt.close()
    
    
def temperature_statistic(T,FID,weight_bool=False,savefile=None):
    """
    Plot of temperature distribution including some statistics about the temperature map.
    Input
        - T: numpy.ma.core.MaskedArrays (Ts or Ts_fixed)
        - FID: str
        - weight_bool: boolean (if histogram should be weighted) 
        - savefile: str
    Output
        - Figure will be created
        - stat_dict: dictonary 
    """
    
    T_flat = T[~np.isnan(T)] # exclude gird cells without values
    bins = np.arange(np.nanmin(T), np.nanmax(T),.02) # 0.02 is the temperature accuracy of the IR camera
    
    plt.figure(figsize=(12,9))
    if weight_bool:
        plt.hist(T_flat, bins=bins, histtype='step', color='k', weights=np.ones_like(T_flat)/len(T_flat))
        plt.ylabel('probability')
    else:
        plt.hist(T_flat, bins=bins, histtype='step', color='k')
        plt.ylabel('frequency')
        
    plt.xlabel('temperature / K')
    plt.title(FID)
    
    if savefile==None:
        print('No path and/or filename. Figure will be not save.')
        plt.show()
    else:
        plt.savefig(savefile)
        plt.close()
        
    stat_dict = {} # dictonary for statistical values
    stat_dict['mean / K'] = round(np.nanmean(T),2)
    stat_dict['median / K'] = round(np.nanmedian(T),2)
    stat_dict['standard deviation / K'] = round(np.nanstd(T),2)
    stat_dict['1st Percentile / K'] = round(np.nanpercentile(T,1),2)
    stat_dict['25th Percentile / K'] = round(np.nanpercentile(T,25),2) 
    stat_dict['75th Percentile / K'] = round(np.nanpercentile(T,75),2) 
    stat_dict['99th Percentile / K'] = round(np.nanpercentile(T,99),2)
    
    return stat_dict


if __name__ == "__main__":
    """
    """