#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This collection contains function for corrections which were applied to the images. It also includes the read function for the image data and the display function for flight parameter.

@author: Linda Thielke
"""

# Required modules
import numpy as np
from netCDF4 import Dataset
import matplotlib
import matplotlib.pyplot as plt

from metadata_heli_flights import get_image_cutoff_flight

# global parameters
PIXEL_WIDTH = 640
PIXEL_HEIGHT = 480

# global setting for figure font
font = {'size': 16}
matplotlib.rc('font', **font)


def load_image_data(loadfile):
    """
    Data reading of gridded data from NetCDF file with netCDF4.Dataset.
    The current output is limited to the relative coordinates and the 
    surface temperature and can be extended indvidually.
   
    Input
       - loadfile: str
    Output
       - lon_img,lat_img,Ts: numpy.ma.core.MaskedArrays
       - FID: str
    """

    data=Dataset(loadfile)
    
    # Flight identification
    devops=data.Device_Operation
    FID=data.Flight_ID
    
    
    # coordinate array for every pixel
    lon_img=data.variables['lon_georef'][:]
    lat_img=data.variables['lat_georef'][:]
    
    # coordinate of nadir point of camera
    lon_nadir=data.variables['lon_nadir'][:]
    lat_nadir=data.variables['lat_nadir'][:]
    
    Ts=data.variables['Ts'][:] # surface temperature
    
    time=data.variables['time'][:] # time in seconds since daystart
    heading=data.variables['head'][:]
    altitude=data.variables['alti'][:]
    
    roll=data.variables['roll'][:] # roll angle in degree
    pitch=data.variables['pitch'][:] # pitch angle in degree

#     print('Projection information:')
#     print(data.variables['stereographic'])
    
    return lon_nadir, lat_nadir, time, altitude, Ts, FID, devops


def get_mask(FID):
    """
    Get array of images with the corner mask, where the zero values are affected and should be removed.
    Input
        - FID: str (Flight ID)
    Output
        - mask: np.ndarray (binary)
    """
    
    mask = np.ones((480,640)) # initial array
    
    # cases
    ll_corner=['20191020_01', '20191029_01', '20191105_01', 
               '20191112_01', '20191112_02', '20191119_01'] # lower left corner
    no_corner=['20191002_01']
    
    if FID in ll_corner:
        print('Lower left corner mask.')
        mask[:50,:100]=0
    elif FID in no_corner:
        print('No corner mask required.')
        pass
    else: # most cases
        print('Upper right corner mask.')
        mask[-50:,-100:]=0
        
    return mask


def get_image_gradient_correction(T_imgs):
    """
    For operational use, included in the function.
    - input:
        T_imgs: numpy.ma.core.MaskedArray of numpy.ma.core.MaskedArrays (series of temperature images)
    - output:type()
        T_grad: numpy.ndarray
    """
    
    T_diff=np.zeros((480,640)) # intial difference array
    count_interation=0 # initial counting variable
    
    for T in T_imgs:
        # summation of relative difference to the center point of the image
        T_diff += T/T[int(PIXEL_HEIGHT/2),int(PIXEL_WIDTH/2)]
        count_interation+=1
        
    T_grad = T_diff/count_interation # average difference of every pixel
    
    print('Number of images used for empirical gradient: ', count_interation) 
    
    return T_grad


def plot_flight_parameter(lon_nadir,lat_nadir,time,alti,FID,devops,savefile=None):
    """
    """
    
    # values for cut-off at start and end of flight
    cutoff = get_image_cutoff_flight()
    cutoff = cutoff[devops] 
    
    # Flight track
    fig, (ax1,ax2) = plt.subplots(2,1,figsize=(12,15))
    ax1.plot(lon_nadir, lat_nadir, color='k')
    ax1.plot(lon_nadir[cutoff[0]:-cutoff[1]], lat_nadir[cutoff[0]:-cutoff[1]], color='r', label='cutoff')
    ax1.set_xlabel('longitude / east degrees')
    ax1.set_ylabel('latitude / north degrees')
    ax1.legend()
    ax1.set_title(FID)

    # Flight Altitude
    ax2.plot(time, alti, color='k')
    ax2.plot(time[cutoff[0]:-cutoff[1]], alti[cutoff[0]:-cutoff[1]], color='r', label='cutoff')
    ax2.set_xlabel('time / seconds since daystart')
    ax2.set_ylabel('flight altitude / m')
    ax2.legend()
    ax2.set_title(FID)
    
    if savefile==None:
        print('No path and/or filename. Figure will be not save.')
        plt.show()
    else:
        plt.savefig(savefile)
        plt.close()


if __name__ == "__main__":
    """
    """