#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Here we present the functions to calculate the time-fixed surface temperature bases on the surface temperature.
It includes four fit functions which are used for correctiopn based on the image data which is later applied to the gridded data.

@author: Linda Thielke
"""

# Required modules
import numpy as np
from pyproj import Proj
from scipy.optimize import curve_fit
from scipy.stats import chisquare

def linear(x, a1, c):
    """
    Linear fit function.
    """
    return a1 * x + c


def quadratic(x, a1, a2, c):
    """
    Quadratic fit function.
    """
    return a2 * x**2 + a1 * x + c


def poly3(x, a1, a2, a3, c):
    """
    Third degree polynomial fit function.
    """
    return a3 * x**3 + a2 * x**2 + a1 * x + c


def exp(x, a, b, c):
    """
    Exponential fit function.
    """
    return a * np.exp(-b*(x-np.nanmin(x)))+c


def get_center_mask(lon,lat,lon_center,lat_center,devops):
    """
    Creating an individual maks for each flight based on the coordinates and
    a fixed distance to the given center point. Note that the center coordinate
    is usually the RV Polarstern coordinate at the targettime. There are four cases
    where RV PS is not located in the map. For this cases ('PS122-1_2-57','PS122-2_19-46', 
    'PS122-2_19-52', 'PS122-2_19-53') the mean coordinates have to be used.    
    Input
        - lon, lat: numpy.ma.core.MaskedArrays (nadir coordinates of images)
        - lon_center, lat_center: floats (reference coordinates)
        - devops: str (identification for special case)
    Output
        - mask: numpy.array
    """
    
    proj_dict = dict(proj='stere', 
                     lat_0=lat_center, 
                     lat_ts=lat_center,
                     lon_0=lon_center,
                     k=1,
                     x_0=0,
                     y_0=0,
                     a=6378137, 
                     b=6378137,
                     units='m')
    m=Proj(proj_dict)
    
    # get center area
    dist=1000
    if devops=='PS122-2_17-101': # center not covered, extended map flight (of triple big map)
        lon_p,lat_p = m(np.arange(-1*dist-1000,dist-1000),np.arange(-1*dist,dist),inverse=True)
        print('Area shift')
    else:
        lon_p,lat_p = m(np.arange(-1*dist,dist),np.arange(-1*dist,dist),inverse=True)
    
    # Corner points of center area
    lon_min,lon_max,lat_min,lat_max = (np.min(lon_p),np.max(lon_p),np.min(lat_p),np.max(lat_p))
    
    # boolean array for center area
    mask = (lon>lon_min) & (lon<lon_max) & (lat>lat_min) & (lat<lat_max)
    mask[:100]=True # including start (to cover whole flight)
    mask[-100:]=True # including end (to cover whole flight)
    
    return mask


def get_best_fit(time,Ts_imgs, mask, devops):
    """
    Defining the best fit for the data based on the image series from the single image data set.
    Input
        - time_imgs: numpy.ma.core.MaskedArrays
        - Ts_imgs: numpy.ma.core.MaskedArrays
        - mask: numpy.array
    Output
        - fit_tyoe: str 
        - p_*: pandas.DataFrame (fit parameter)
        - chis_all: numpy.array (Chi squared statistics)
    """
    
    # masking to get center area data
    Ts = []
    for Ti in Ts_imgs.data:
        Ts.append(np.quantile(Ti,.1))
    x=time[mask]
    y=np.asarray(Ts)[mask]
    
    # linear fit
    p_lin = curve_fit(linear, x, y)
    y_lin = linear(x,p_lin[0][0],p_lin[0][1])
    y_lin_all = linear(time,p_lin[0][0],p_lin[0][1])

    # quadratic fit
    p_quad = curve_fit(quadratic, x, y)
    y_quad = quadratic(x,p_quad[0][0],p_quad[0][1],p_quad[0][2])
    y_quad_all = quadratic(time,p_quad[0][0],p_quad[0][1],p_quad[0][2])

    # 3rd degree polynomial fit
    p_poly = curve_fit(poly3, x, y)
    y_poly = poly3(x,p_poly[0][0],p_poly[0][1],p_poly[0][2],p_poly[0][3])
    y_poly_all = poly3(time,p_poly[0][0],p_poly[0][1],p_poly[0][2],p_poly[0][3])

    # different initial guesses for different flights required
    if devops in ['PS122-1_7-24','PS122-2_19-45','PS122-2_20-53','PS122-2_21-78']:
        p_exp = curve_fit(exp, x, y, (-0.1,0.0001,200))
    elif devops in ['PS122-1_8-23','PS122-2_17-98', 'PS122-2_19-44', 'PS122-2_19-52','PS122-2_19-53', 
                    'PS122-2_22-97', 'PS122-2_25-7', 'PS122-3_37-63']:
        p_exp = curve_fit(exp, x, y, (0,-0.0001,200))
    else:
        p_exp = curve_fit(exp, x, y, (0,0.0001,200))
    
    # exponential fit
    y_exp = exp(x,p_exp[0][0],p_exp[0][1],p_exp[0][2])
    y_exp_all = exp(time,p_exp[0][0],p_exp[0][1],p_exp[0][2])
    
    # fit testing with Chi squared
    # over masked data only
    chisq_lin = chisquare(y, y_lin)
    chisq_quad = chisquare(y, y_quad)
    chisq_poly = chisquare(y, y_poly)
    chisq_exp = chisquare(y, y_exp)

    # over all data
    chisq_lin_all = chisquare(Ts, y_lin_all)
    chisq_quad_all = chisquare(Ts, y_quad_all)
    chisq_poly_all = chisquare(Ts, y_poly_all)
    chisq_exp_all = chisquare(Ts, y_exp_all)

    chis = [chisq_lin.statistic,chisq_quad.statistic,chisq_poly.statistic,chisq_exp.statistic]
    chis_all = [chisq_lin_all.statistic,chisq_quad_all.statistic,chisq_poly_all.statistic,chisq_exp_all.statistic]
    
    # Minium = best fitting
    min_index = chis_all.index(min(chis_all))
    fit_p = {}
    # Output
    if min_index==0:
        fit_type='linear'
        print(fit_type)
        fit_p['a1']=p_lin[0][0]
        fit_p['x']=p_lin[0][1]
        return fit_type, fit_p, chis_all
    elif min_index==1:
        fit_type='quadratic'
        print(fit_type)
        fit_p['a1']=p_quad[0][0]
        fit_p['a2']=p_quad[0][1]
        fit_p['c']=p_quad[0][2]
        return fit_type, fit_p, chis_all
    elif min_index==2:
        fit_type='polynomial'
        print(fit_type)
        fit_p['a1']=p_poly[0][0]
        fit_p['a2']=p_poly[0][1]
        fit_p['a3']=p_poly[0][2]
        fit_p['c']=p_poly[0][3]
        return fit_type, fit_p, chis_all
    elif min_index==3:
        fit_type='exponential'
        print(fit_type)
        fit_p['a']=p_exp[0][0]
        fit_p['b']=p_exp[0][1]
        fit_p['c']=p_exp[0][2]
        return fit_type, fit_p, chis_all
    
    
def get_timefixed_Ts(targettime, time_map, Ts_map, fit_type, fit_param):
    """
    Correcting the gridded surface temperature based on the selected fit type and
    the calculated fit parameter. 
    Input
        - targettime: float (in seconds since daystart)
        - time_map: numpy.ma.core.MaskedArrays
        - Ts_map: numpy.ma.core.MaskedArrays
        - fit_type: str
        - fit_param: dictonary
    """
    
    # y: fit over the whole map
    # y_0: reference value of fit at the targettime
    if fit_type=='linear':
        y = linear(time_map,fit_param['a1'],fit_param['c'])
        y_0 = linear(targettime,fit_param['a1'],fit_param['c'])
    elif fit_type=='quadratic':
        y = quadratic(time_map,fit_param['a1'],fit_param['a2'],fit_param['c'])
        y_0 = quadratic(targettime,fit_param['a1'],fit_param['a2'],fit_param['c'])
    elif fit_type=='polynomial':
        y = poly3(time_map,fit['a1'],fit_param['a2'],fit_param['a3'],fit_param['c'])
        y_0 = poly3(targettime,fit['a1'],fit_param['a2'],fit_param['a3'],fit_param['c'])
    elif fit_type=='exponential':
        y = exp(time_map,fit_param['a'],fit_param['b'],fit_param['c'])
        y_0 = exp(targettime,fit_param['a'],fit_param['b'],fit_param['c'])

    Ts_fixed = Ts_map - (y - y_0) # corrected gridded data
    
    return Ts_fixed


if __name__ == "__main__":
    """
    """
    
#     # --- possible function calls:
#     import image_data_corrections as idc
#     import gridded_data_reading_plotting as gd
    
#     xc, yc, lon, lat, time, Ts, FID = gd.read_map_nc(file_map)
#     lon_nadir, lat_nadir, time_img, altitude, Ts_img, FID, devops = idc.load_image_data(file_img)
    
#     mask = tfc.get_center_mask(lon_nadir,lat_nadir,np.mean(lon),np.mean(lat),devops)
#     fit_type, fit_param, chis_stat_all = tfc.get_best_fit(time_img,Ts_img, mask, devops)
    
#     Ts_fixed_map = tfc.get_timefixed_Ts(targettime, time, Ts, fit_type, fit_param)